﻿using System;
using System.ComponentModel;
using System.Diagnostics;
using System.IO;
using System.Security.Cryptography;
using System.ServiceProcess;
using Inedo.Consolation;
using Inedo.WindowsServices;

namespace Inedo.Agents.Service
{
    public static class Program
    {
        [ConsoleCommand]
        [Description("Runs the agent service interactively.")]
        public static void run([ConsoleAlias("s"), Browsable(false)] bool runAsService = false)
        {
            if (runAsService)
            {
                ServiceBase.Run(AgentService.Instance.GetWindowsService());
            }
            else
            {
                ConsoleCancelEventHandler cancel =
                    (s, e) =>
                    {
                        e.Cancel = true;
                        AgentService.Instance.Stop();
                    };

                AgentService.Instance.Start();
                Console.CancelKeyPress += cancel;

                try
                {
                    AgentService.Instance.Task.Wait();
                }
                finally
                {
                    Console.CancelKeyPress -= cancel;
                }
            }
        }

        [ConsoleCommand]
        [Description("Installs the agent service as a Windows service.")]
        public static void install([Description("The user name of the service account.")] string user = null, [Description("The password of the service account.")] string password = null)
        {
            using (var service = WindowsService.GetService(AgentService.Instance.Name))
            {
                if (service != null)
                {
                    Console.WriteLine($"Service {AgentService.Instance.Name} already exists; attempting to uninstall...");
                    service.Delete();
                    Console.WriteLine("Service removed.");
                }
            }

            var commandLine = $"\"{typeof(Program).Assembly.Location}\" run -s";

            Console.WriteLine($"Attempting to create service {AgentService.Instance.Name}...");
            using (var service = WindowsService.CreateService(AgentService.Instance.Name, AgentService.Instance.FriendlyName, commandLine, user, password))
            {
                service.Description = AgentService.Instance.Description;
                Console.WriteLine("Service installed.");
            }
        }

        [ConsoleCommand]
        [Description("Uninstalls the agent Windows service.")]
        public static void uninstall()
        {
            using (var service = WindowsService.GetService(AgentService.Instance.Name))
            {
                if (service != null)
                {
                    Console.WriteLine($"Uninstalling service {AgentService.Instance.Name}...");
                    service.Delete();
                    Console.WriteLine("Service removed.");
                }
                else
                {
                    Console.WriteLine($"Service {AgentService.Instance.Name} is not installed.");
                }
            }
        }

        [ConsoleCommand]
        [Description("Generates a random AES encryption key.")]
        public static void generateKey()
        {
            using (var random = new RNGCryptoServiceProvider())
            {
                var data = new byte[16];
                random.GetBytes(data);
                Console.WriteLine(AgentEndpoint.FormatEncryptionKey(data));
            }
        }

        [ConsoleCommand]
        [Browsable(false)]
        public static void update(string targetPath, string serviceName, int pid)
        {
            using (var serviceProcess = Process.GetProcessById(pid))
            {
                try
                {
                    using (var service = new ServiceController(serviceName))
                    {
                        service.Stop();
                        service.WaitForStatus(ServiceControllerStatus.Stopped);
                    }
                }
                catch
                {
                }

                if (!serviceProcess.HasExited)
                {
                    if (!serviceProcess.WaitForExit(5000))
                    {
                        serviceProcess.Kill();
                    }
                }
            }

            foreach (var sourceFile in Directory.EnumerateFiles(Path.GetDirectoryName(typeof(Program).Assembly.Location)))
            {
                var targetFile = Path.Combine(targetPath, Path.GetFileName(sourceFile));
                File.Copy(sourceFile, targetFile, true);
            }

            using (var service = new ServiceController(serviceName))
            {
                service.Start();
            }
        }

        public static int Main(string[] args)
        {
            return ConsoleApplication.Run(typeof(Program), args);
        }
    }
}
